import re
import asyncio
from pathlib import Path
from dataclasses import dataclass

from adb_shell.adb_device_async import AdbDeviceTcpAsync
from adb_shell.auth.sign_pythonrsa import PythonRSASigner
from adb_shell.auth.keygen import keygen
from adb_shell.exceptions import TcpTimeoutException

from googletv.constants import (
    ADB_MAX_RETRIES,
    ADB_RETRY_INTERVAL,
    ADB_MAX_TIMEOUTS,
    APP_ATV_LAUNCHER,
    APP_NETFLIX,
    APP_NLZIET,
    APP_PLEX,
    APP_TVHEADEND,
    APP_VLC,
    APP_YOUTUBE,
    APP_NAMES,
    CMD_DUMPSYS_ACTIVITY,
    CMD_DUMPSYS_DISPLAY,
    CMD_DUMPSYS_ERROR,
    CMD_DUMPSYS_MEDIA_SESSION,
    CMD_DUMPSYS_POWER,
    REGEX_ACTIVITY_RESUMED,
    REGEX_ACTIVITY,
    REGEX_ASLEEP,
    REGEX_AWAKE,
    REGEX_PLAYBACK_STATE,
    REGEX_SCREEN_OFF,
    REGEX_SCREEN_ON,
    REGEX_SESSION,
    REGEX_WAKE_LOCK_SIZE,
    STATE_IDLE,
    STATE_OFF,
    STATE_PAUSED,
    STATE_PLAYING,
)
from googletv.exception import GoogleTvException
from googletv.search import search_nested


class AdbKey:
    def __init__(self, file_path: str) -> None:
        self._file_path = file_path

        self._public_key: str | None = None
        self._private_key: str | None = None

    @property
    def public_key_path(self) -> str:
        return f"{self._file_path}.pub"

    @property
    def private_key_path(self) -> str:
        return self._file_path

    @property
    def public_key(self) -> str:
        if self._public_key is None:
            with open(f"{self._file_path}.pub", "r") as f:
                self._public_key = f.read()

        return self._public_key

    @property
    def private_key(self) -> str:
        if self._private_key is None:
            with open(self._file_path, "r") as f:
                self._private_key = f.read()

        return self._private_key

    @property
    def exists(self) -> bool:
        for path in (self.private_key_path, self.public_key_path):
            if not Path(path).is_file():
                return False

        return True

    @staticmethod
    def generate(file_path: str) -> "AdbKey":
        keygen(file_path)
        return AdbKey(file_path)


@dataclass(frozen=True)
class DeviceState:
    awake: bool | None = None
    screen_on: bool | None = None
    app: str | None = None
    playback_state: str | None = None


class GoogleTv:
    def __init__(self, adb_key: AdbKey, host: str, port: int = 5555) -> None:
        self._adb = AdbDeviceTcpAsync(host, port)
        self._adb_key = adb_key

        self._state = DeviceState()
        self._timeout_counter = 0

    @property
    def available(self) -> bool:
        return self._adb.available

    @property
    def state(self) -> DeviceState:
        return self._state

    async def connect(self) -> None:
        signer = PythonRSASigner(self._adb_key.public_key, self._adb_key.private_key)
        await self._adb.connect(rsa_keys=[signer])

    async def close(self) -> None:
        await self._adb.close()

    async def reconnect(self) -> None:
        await self.close()
        await self.connect()

    async def update(self) -> None:
        dumpsys_power = await self._send_command(CMD_DUMPSYS_POWER)
        awake = self._parse_awake(dumpsys_power)
        if not awake:
            new_state = DeviceState(
                awake=awake,
                screen_on=False,
                playback_state=STATE_OFF,
            )
        else:
            dumpsys_display = await self._send_command(CMD_DUMPSYS_DISPLAY)
            dumpsys_activity = await self._send_command(CMD_DUMPSYS_ACTIVITY)
            dumpsys_media_session = await self._send_command(CMD_DUMPSYS_MEDIA_SESSION)

            app = self._parse_app(dumpsys_activity)
            if app:
                playback_state = self._parse_playback_state(
                    dumpsys_power, dumpsys_media_session, app
                )
            else:
                playback_state = None

            app_name = APP_NAMES.get(app, app) if app else None

            new_state = DeviceState(
                awake=self._parse_awake(dumpsys_power),
                screen_on=self._parse_screen_on(dumpsys_display),
                app=app_name,
                playback_state=playback_state,
            )

        self._state = new_state

    async def _send_command(
        self,
        command: str,
        retries: int = ADB_MAX_RETRIES,
        retry_interval: float = ADB_RETRY_INTERVAL,
    ) -> str:
        for _ in range(retries):
            if self._timeout_counter >= ADB_MAX_TIMEOUTS:
                await self.reconnect()

            try:
                result = await self._adb.shell(command)
            except TcpTimeoutException:
                self._timeout_counter += 1
                await asyncio.sleep(retry_interval)
                continue

            self._timeout_counter = 0

            if CMD_DUMPSYS_ERROR not in result:
                return result

            await asyncio.sleep(retry_interval)

        raise GoogleTvException(f"Failed to send command {command}")

    def _parse_awake(self, dumpsys_power: str) -> bool | None:
        if re.search(REGEX_AWAKE, dumpsys_power):
            return True
        if re.search(REGEX_ASLEEP, dumpsys_power):
            return False
        return None

    def _parse_screen_on(self, dumpsys_display: str) -> bool | None:
        if re.search(REGEX_SCREEN_ON, dumpsys_display):
            return True
        if re.search(REGEX_SCREEN_OFF, dumpsys_display):
            return False
        return None

    def _parse_app(self, dumpsys_activity: str) -> str | None:
        matches = search_nested(
            (REGEX_ACTIVITY, REGEX_ACTIVITY_RESUMED), dumpsys_activity
        )
        if matches:
            return matches[0].group("app")

        return None

    def _parse_playback_state(
        self, dumpsys_power: str, dumpsys_media_session: str, app: str
    ) -> str | None:
        wl_match = re.search(REGEX_WAKE_LOCK_SIZE, dumpsys_power)
        if not wl_match:
            return None

        ms_matches = search_nested(
            (REGEX_SESSION, app, REGEX_PLAYBACK_STATE), dumpsys_media_session
        )
        if not ms_matches:
            return None

        wake_lock_size = int(wl_match.group("wake_lock_size"))
        media_session_state = int(ms_matches[2].group("state"))

        # ATV Launcher
        if app in [APP_ATV_LAUNCHER, None]:
            return STATE_IDLE

        # Netflix
        elif app == APP_NETFLIX:
            if media_session_state == 2:
                return STATE_PAUSED
            elif media_session_state == 3:
                return STATE_PLAYING
            else:
                return STATE_IDLE

        # NLZIET
        elif app == APP_NLZIET:
            if wake_lock_size == 1:
                return STATE_PAUSED
            elif wake_lock_size == 2:
                return STATE_PLAYING
        # Plex
        elif app == APP_PLEX:
            if media_session_state == 3:
                if wake_lock_size == 1:
                    return STATE_PAUSED
                else:
                    return STATE_PLAYING
            else:
                return STATE_IDLE

        # TVheadend
        elif app == APP_TVHEADEND:
            if wake_lock_size == 5:
                return STATE_PAUSED
            elif wake_lock_size == 6:
                return STATE_PLAYING
            else:
                return STATE_IDLE

        # VLC
        elif app == APP_VLC:
            if media_session_state == 2:
                return STATE_PAUSED
            elif media_session_state == 3:
                return STATE_PLAYING
            else:
                return STATE_IDLE

        # YouTube
        elif app == APP_YOUTUBE:
            if media_session_state == 2:
                return STATE_PAUSED
            elif media_session_state == 3:
                return STATE_PLAYING
            else:
                return STATE_IDLE

        # Get the state from `media_session_state`
        elif media_session_state:
            if media_session_state == 2:
                return STATE_PAUSED
            elif media_session_state == 3:
                return STATE_PLAYING
            else:
                return STATE_IDLE

        # Get the state from `wake_lock_size`
        else:
            if wake_lock_size == 1:
                return STATE_PAUSED
            elif wake_lock_size == 2:
                return STATE_PLAYING
            else:
                return STATE_IDLE

        return None



ADB_MAX_RETRIES = 3
ADB_RETRY_INTERVAL = 0.25
ADB_MAX_TIMEOUTS = 2

CMD_DUMPSYS_DISPLAY = "dumpsys display"
CMD_DUMPSYS_POWER = "dumpsys power"
CMD_DUMPSYS_ACTIVITY = "dumpsys activity ."
CMD_DUMPSYS_MEDIA_SESSION = "dumpsys media_session"

CMD_DUMPSYS_ERROR = "Error dumping service info"

REGEX_AWAKE = r"mWakefulness=Awake|Dreaming"
REGEX_ASLEEP = r"mWakefulness=Asleep"

REGEX_SCREEN_ON = r"mScreenState=ON"
REGEX_SCREEN_OFF = r"mScreenState=OFF"

REGEX_WAKE_LOCK_SIZE = r"Wake Locks: size=(?P<wake_lock_size>\d+)"

REGEX_ACTIVITY = r"ACTIVITY (?P<app>[^ ]+)\/(?P<activity>[^ ]+)"
REGEX_ACTIVITY_RESUMED = r"mResumed=true"

REGEX_SESSION = r"Sessions Stack"
REGEX_PLAYBACK_STATE = r"state=PlaybackState {state=(?P<state>\d)"

# Playback states

STATE_OFF = "off"
STATE_IDLE = "idle"
STATE_PLAYING = "playing"
STATE_PAUSED = "paused"

# Apps
APP_AE_TV = "com.aetn.aetv.watch"
APP_AMAZON_PRIME_VIDEO = "com.amazon.avod.thirdpartyclient"
APP_AMAZON_VIDEO = "com.amazon.avod"
APP_APPLE_TV_PLUS = "com.apple.atve.androidtv.appletv"
APP_ATV_LAUNCHER = "com.google.android.tvlauncher"
APP_BELL_FIBE = "com.quickplay.android.bellmediaplayer"
APP_CBC_GEM = "ca.cbc.android.cbctv"
APP_COMEDY_CENTRAL = "com.vmn.android.comedycentral"
APP_CRAVE = "ca.bellmedia.cravetv"
APP_DAILYMOTION = "com.dailymotion.dailymotion"
APP_DEEZER = "deezer.android.tv"
APP_DISNEY_PLUS = "com.disney.disneyplus"
APP_DISNEY_PLUS_HOTSTAR = "in.startv.hotstar"
APP_DS_PHOTO = "com.synology.dsphoto"
APP_DS_VIDEO = "com.synology.dsvideo"
APP_ES_FILE_EXPLORER = "com.estrongs.android.pop"
APP_FACEBOOK = "com.facebook.katana"
APP_FAWESOME = "com.future.moviesByFawesomeAndroidTV"
APP_FIREFOX = "org.mozilla.tv.firefox"
APP_FOOD_NETWORK_GO = "tv.accedo.foodnetwork"
APP_FRANCE_TV = "fr.francetv.pluzz"
APP_GLOBAL_TV = "com.shawmedia.smglobal"
APP_GOOGLE_CAST = "com.google.android.apps.mediashell"
APP_GOOGLE_TV_LAUNCHER = "com.google.android.apps.tv.launcherx"
APP_HAYSTACK_NEWS = "com.haystack.android"
APP_HBO_GO = "eu.hbogo.androidtv.production"
APP_HBO_GO_2 = "com.HBO"
APP_HOICHOI = "com.viewlift.hoichoi"
APP_HULU = "com.hulu.plus"
APP_HUNGAMA_PLAY = "com.hungama.movies.tv"
APP_IMDB_TV = "com.amazon.imdb.tv.android.app"
APP_IPTV = "ru.iptvremote.android.iptv"
APP_IPTV_SMARTERS_PRO = "com.nst.iptvsmarterstvbox"
APP_JELLYFIN_TV = "org.jellyfin.androidtv"
APP_JIO_CINEMA = "com.jio.media.stb.ondemand"
APP_KODI = "org.xbmc.kodi"
APP_LIVE_CHANNELS = "com.google.android.tv"
APP_MIJN_RADIO = "org.samsonsen.nederlandse.radio.holland.nl"
APP_MOLOTOV = "tv.molotov.app"
APP_MRMC = "tv.mrmc.mrmc"
APP_MRMC_LITE = "tv.mrmc.mrmc.lite"
APP_MX_PLAYER = "com.mxtech.videoplayer.ad"
APP_NETFLIX = "com.netflix.ninja"
APP_NLZIET = "nl.nlziet"
APP_NOS = "nl.nos.app"
APP_NPO = "nl.uitzendinggemist"
APP_OCS = "com.orange.ocsgo"
APP_PLAY_GAMES = "com.google.android.play.games"
APP_PLAY_MUSIC = "com.google.android.music"
APP_PLAY_STORE = "com.android.vending"
APP_PLAY_VIDEOS = "com.google.android.videos"
APP_PLEX = "com.plexapp.android"
APP_PRIME_VIDEO = "com.amazon.amazonvideo.livingroom"
APP_SETTINGS = "com.android.tv.settings"
APP_SMART_YOUTUBE_TV = "com.liskovsoft.videomanager"
APP_SPORT1 = "de.sport1.firetv.video"
APP_SPOTIFY = "com.spotify.tv.android"
APP_STEAM_LINK = "com.valvesoftware.steamlink"
APP_SYFY = "com.amazon.webapps.nbc.syfy"
APP_T2 = "tv.perception.clients.tv.android"
APP_TED = "com.ted.android.tv"
APP_TUNEIN = "tunein.player"
APP_TVHEADEND = "de.cyberdream.dreamepg.tvh.tv.player"
APP_TWITCH = "tv.twitch.android.app"
APP_VEVO = "com.vevo.tv"
APP_VH1 = "com.mtvn.vh1android"
APP_VIMEO = "com.vimeo.android.videoapp"
APP_VLC = "org.videolan.vlc"
APP_VOYO = "com.phonegap.voyo"
APP_VRV = "com.ellation.vrv"
APP_WAIPU_TV = "de.exaring.waipu.firetv.live"
APP_WATCH_TNT = "com.turner.tnt.android.networkapp"
APP_YOUTUBE = "com.google.android.youtube.tv"
APP_YOUTUBE_KIDS = "com.google.android.youtube.tvkids"
APP_YOUTUBE_MUSIC = "com.google.android.youtube.tvmusic"
APP_YOUTUBE_TV = "com.google.android.youtube.tvunplugged"
APP_ZEE5 = "com.graymatrix.did"
APP_ZIGGO_GO_TV = "com.ziggo.tv"
APP_NAMES = {
    APP_AE_TV: "A&E",
    APP_AMAZON_PRIME_VIDEO: "Amazon Prime Video",
    APP_AMAZON_VIDEO: "Amazon Video",
    APP_APPLE_TV_PLUS: "Apple TV+",
    APP_ATV_LAUNCHER: "Android TV Launcher",
    APP_BELL_FIBE: "Bell Fibe",
    APP_CBC_GEM: "CBC Gem",
    APP_COMEDY_CENTRAL: "Comedy Central",
    APP_CRAVE: "Crave",
    APP_DAILYMOTION: "Dailymotion",
    APP_DEEZER: "Deezer",
    APP_DISNEY_PLUS: "Disney+",
    APP_DISNEY_PLUS_HOTSTAR: "Disney+ Hotstar",
    APP_DS_PHOTO: "DS photo",
    APP_DS_VIDEO: "DS video",
    APP_ES_FILE_EXPLORER: "ES File Explorer",
    APP_FACEBOOK: "Facebook Watch",
    APP_FAWESOME: "Fawsome",
    APP_FIREFOX: "Firefox",
    APP_FOOD_NETWORK_GO: "Food Network GO",
    APP_FRANCE_TV: "France TV",
    APP_GLOBAL_TV: "Global TV",
    APP_GOOGLE_CAST: "Google Cast",
    APP_GOOGLE_TV_LAUNCHER: "Google TV Launcher",
    APP_HAYSTACK_NEWS: "Haystack News",
    APP_HBO_GO: "HBO GO",
    APP_HBO_GO_2: "HBO GO (2)",
    APP_HOICHOI: "Hoichoi",
    APP_HULU: "Hulu",
    APP_HUNGAMA_PLAY: "Hungama Play",
    APP_IMDB_TV: "IMDb TV",
    APP_IPTV: "IPTV",
    APP_IPTV_SMARTERS_PRO: "IPTV Smarters Pro",
    APP_JELLYFIN_TV: "Jellyfin",
    APP_JIO_CINEMA: "Jio Cinema",
    APP_KODI: "Kodi",
    APP_LIVE_CHANNELS: "Live Channels",
    APP_MIJN_RADIO: "Mijn Radio",
    APP_MOLOTOV: "Molotov",
    APP_MRMC: "MrMC",
    APP_MRMC_LITE: "MrMC Lite",
    APP_MX_PLAYER: "MX Player",
    APP_NETFLIX: "Netflix",
    APP_NLZIET: "NLZIET",
    APP_NOS: "NOS",
    APP_NPO: "NPO",
    APP_OCS: "OCS",
    APP_PLAY_GAMES: "Play Games",
    APP_PLAY_MUSIC: "Play Music",
    APP_PLAY_STORE: "Play Store",
    APP_PLAY_VIDEOS: "Play Movies & TV",
    APP_PLEX: "Plex",
    APP_PRIME_VIDEO: "Prime Video",
    APP_SETTINGS: "Settings",
    APP_SMART_YOUTUBE_TV: "Smart YouTube TV",
    APP_SPORT1: "Sport 1",
    APP_SPOTIFY: "Spotify",
    APP_STEAM_LINK: "Steam Link",
    APP_SYFY: "Syfy",
    APP_T2: "T-2 TV",
    APP_TED: "TED",
    APP_TUNEIN: "TuneIn Radio",
    APP_TVHEADEND: "DreamPlayer TVHeadend",
    APP_TWITCH: "Twitch",
    APP_VEVO: "Vevo",
    APP_VH1: "VH1",
    APP_VIMEO: "Vimeo",
    APP_VLC: "VLC",
    APP_VOYO: "VOYO",
    APP_VRV: "VRV",
    APP_WAIPU_TV: "Waipu TV",
    APP_WATCH_TNT: "Watch TNT",
    APP_YOUTUBE: "YouTube",
    APP_YOUTUBE_KIDS: "YouTube Kids",
    APP_YOUTUBE_MUSIC: "YouTube Music",
    APP_YOUTUBE_TV: "YouTube TV",
    APP_ZEE5: "ZEE5",
    APP_ZIGGO_GO_TV: "Ziggo GO TV",
}



import re
from typing import Sequence


def search_nested(
    patterns: Sequence[str], string: str, flags: int = 0
) -> Sequence[re.Match[str]] | None:
    matches = list(re.finditer(patterns[0], string, flags))
    for i, match in enumerate(matches):
        if len(patterns) == 1:
            return (match,)

        start = match.start()
        try:
            end = matches[i + 1].start()
        except IndexError:
            end = len(string)

        nested_match = search_nested(patterns[1:], string[start:end], flags)
        if nested_match:
            return (match, *nested_match)

    return None



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['googletv']

package_data = \
{'': ['*']}

install_requires = \
['adb-shell[async]>=0.4.3,<0.5.0']

setup_kwargs = {
    'name': 'googletv',
    'version': '0.0.2',
    'description': '',
    'long_description': '# python-googletv',
    'author': 'Sören Oldag',
    'author_email': 'soeren_oldag@freenet.de',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10',
}


setup(**setup_kwargs)
